<?php

/**
 * Class JotUrlSDK
 */
class JotUrlSDK {

	const USER_AGENT = 'JotUrl SDK 1.2.PHP (https://www.joturl.com/)';

	#base URL for endpoints
	private static $baseURL = 'https://joturl.com/a/i1/';

	/**
	 * @var string username to be used in the login
	 */
	private $username;

	/**
	 * @var string ID of the login session
	 */
	private $session_id = '';

	/**
	 * @var string private key, https://www.joturl.com/reserved/settings.html#tools-api
	 */
	private $private_key;

	/**
	 * @var string public key, https://www.joturl.com/reserved/settings.html#tools-api
	 */
	private $public_key;

	/**
	 * @var bool if true shows error on console
	 */
	private $return_errors = false;

	/**
	 * @var string last error
	 */
	private $last_error = '';

	/**
	 * @var array last status
	 */
	private $last_status = array();

	/**
	 * @var string API token
	 */
	private $api_token;


	private $debug = false;

	private $stats = array(
		'count' => 0,
		'endpoints' => array(),
	);

	/**
	 * Creates an instance of the JotUrl SDK.
	 *
	 * @param string      $username               the username used to login into JotURL dashboard
	 * @param string      $public_key_or_rw_token public api key or RW api token, you can find it on https://www.joturl.com/reserved/settings.html#tools-api
	 * @param string|null $private_key            private api key, you can find it on https://www.joturl.com/reserved/settings.html#tools-api
	 */
	public function __construct( $username, $public_key_or_rw_token, $private_key = null ) {
		$this->username = $username;
		if ( empty( $private_key ) ) {
			$this->api_token = $public_key_or_rw_token;
		} else {
			$this->public_key = $public_key_or_rw_token;
			$this->private_key = $private_key;
		}
	}

	public function setDebug( $d = true ) {
		$this->debug = $d;
	}

	public function getStats() {
		ksort( $this->stats['endpoints'] );

		return $this->stats;
	}

	private function setLastError( $error = '' ) {
		$this->last_error = $error;
		if ( ! $this->return_errors ) {
			throw new Exception( $error );
		}
	}

	function setAPIToken( $token = null ) {
		$this->api_token = $token;
	}

	private function resetLastError() {
		$this->last_error = '';
	}

	/**
	 * Returns last occurred error.
	 *
	 * @return string
	 */
	function getLastError() {
		return $this->last_error;
	}

	private function setLastStatus( $status ) {
		$this->last_status = $status;
	}

	private function resetLastStatus() {
		$this->last_status = array();
	}

	/**
	 * Returns last occurred error.
	 *
	 * @return array
	 */
	function getLastStatus() {
		return $this->last_status;
	}

	private function rcall( $url, $postParameters = array(), $retry = true ) {
		$this->resetLastError();
		$this->resetLastStatus();

		try {
			$json = $this->download( $url, $postParameters );
			if ( ! empty( $json ) ) {
				$obj = json_decode( $json, true );

				if ( is_array( $obj ) ) {
					$status = $this->getParam( $obj, 'status', array() );

					$this->setLastStatus( $status );

					$code = (int) $this->getParam( $status, 'code' );
					if ( $code !== 200 ) {
						$this->setLastError( $this->getParam( $status, 'error' ) . ' (CODE=' . $this->getParam( $status, 'code' ) . ' - STATUS=' . $this->getParam( $status, 'text' ) . ')' );

						return false;
					}

					return $this->getParam( $obj, 'result', array() );
				}
			}

			if ( $retry ) {
				#maybe we are calling APIs too fastly
				sleep( 3 );
				if ( $this->login() ) {
					return $this->rcall( $url, $postParameters, false );
				}

				$this->setLastError( 'malformed response (mr2)' );

				return false;
			}
		} catch ( Throwable $t ) {
			$this->setLastError( $t->getMessage() );
		}

		return false;
	}

	/**
	 * Call and get results from the API endpoint.
	 *
	 * @param string $url            URL to be called
	 * @param array  $postParameters [OPTIONAL] array containing post parameters
	 *
	 * @return bool|array associative array containing the result of the call
	 */
	public function call( $url, $postParameters = array(), $retry = true ) {
		return $this->rcall( $url, $postParameters, $retry );
	}

	private function build_data_files( $delimiter, $fields, $files ) {
		$data = '';
		$eol = "\r\n";

		foreach ( $fields as $name => $content ) {
			$data .= "--" . $delimiter . $eol
			         . 'Content-Disposition: form-data; name="' . str_replace( '"', '', $name ) . '"' . $eol . $eol
			         . $content . $eol;
		}

		foreach ( $files as $name => $file ) {
			$data .= "--" . $delimiter . $eol
			         . 'Content-Disposition: form-data; name="' . str_replace( '"', '', $name ) . '"; filename="' . $file->postname . '"' . $eol
			         . 'Content-Type: ' . $file->mime . $eol
			         . 'Content-Transfer-Encoding: binary' . $eol;

			$data .= $eol;
			$data .= file_get_contents( $file->name ) . $eol;
		}
		$data .= "--" . $delimiter . "--" . $eol;


		return $data;
	}

	private function download( $url, $postParameters = array(), $args = array() ) {
		$caCert = __DIR__ . DIRECTORY_SEPARATOR . 'cacert.pem';

		$heads = array();
		$heads[] = 'Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8';
		$heads[] = 'Accept-Encoding: gzip,deflate';
		$heads[] = 'Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7';
		$heads[] = 'Keep-Alive: 115';
		$heads[] = 'Connection: keep-alive';

		if ( ! empty( $this->api_token ) ) {
			$heads[] = 'Authorization: Bearer ' . $this->api_token;
		}

		if ( array_key_exists( 'headers', $args ) && is_array( $args['headers'] ) ) {
			foreach ( $args['headers'] as $ap ) {
				if ( ! in_array( $ap, $heads, true ) ) {
					$heads[] = $ap;
				}
			}
		}

		if ( $this->debug ) {
			$p = rtrim( parse_url( $url, PHP_URL_PATH ), '/' );
			if ( ! array_key_exists( $p, $this->stats['endpoints'] ) ) {
				$this->stats['endpoints'][ $p ] = 0;
			}
			$this->stats['endpoints'][ $p ] ++;
			$this->stats['count'] ++;
		}

		if ( function_exists( 'curl_init' ) ) {
			$args = array_merge(
				array(
					'timeout' => '300',
					'user-agent' => self::USER_AGENT,
					'cookies' => '',
					'redirection' => 5,
					'params' => '',
				), $args
			);

			$options = array(
				CURLOPT_TIMEOUT => $args['timeout'],
				CURLOPT_USERAGENT => $args['user-agent'],
				CURLOPT_RETURNTRANSFER => true,  // return web page
				CURLOPT_HEADER => true,  // return headers
				CURLOPT_FOLLOWLOCATION => true,  // follow redirects
				CURLOPT_ENCODING => "",  // handle all encodings
				CURLOPT_CONNECTTIMEOUT => 120,  // timeout on connect
				CURLOPT_CAINFO => $caCert,
			);

			if ( ! empty( $postParameters ) && is_array( $postParameters ) ) {
				$isMultipart = false;
				$fields = $files = array();
				foreach ( $postParameters as $key => $value ) {
					if ( $value instanceof \CURLFile ) {
						$files[ $key ] = $value;
						$isMultipart = true;
					} else {
						$fields[ $key ] = $value;
					}
				}

				if ( $isMultipart ) {
					$boundary = uniqid( '', true );
					$delimiter = '-------------' . $boundary;

					$post_data = $this->build_data_files( $delimiter, $fields, $files );

					$heads[] = 'Content-Type: multipart/form-data; boundary=' . $delimiter;
					$heads[] = 'Content-Length: ' . strlen( $post_data );

					$options[ CURLOPT_POST ] = 1;
					$options[ CURLOPT_POSTFIELDS ] = $post_data;
				} else {
					$options[ CURLOPT_POST ] = 1;
					$options[ CURLOPT_POSTFIELDS ] = http_build_query( $postParameters );
				}
			}

			$options[ CURLOPT_HTTPHEADER ] = $heads;

			$curl = curl_init( $url );
			curl_setopt_array( $curl, $options );
			$response = curl_exec( $curl );
			$error_number = curl_errno( $curl );
			$error_message = trim( curl_error( $curl ) );
			$http_code = curl_getinfo( $curl, CURLINFO_HTTP_CODE );
			$headerSize = curl_getinfo( $curl, CURLINFO_HEADER_SIZE );
			$body = substr( $response, $headerSize );
			curl_close( $curl );
		} else {
			$error_number = 'FGCE';

			$heads[] = 'User-Agent: ' . self::USER_AGENT;

			$options = array(
				'http' => array(
					'header' => implode( "\r\n", $heads ),
					'method' => 'GET',
					'ignore_errors' => true,
				),
				'ssl' => array(
					'cafile' => $caCert,
				),
			);

			$context = stream_context_create( $options );

			$body = file_get_contents( $url, false, $context );
			$rheaders = isset( $http_response_header ) ? $http_response_header : array( '' );

			$status_line = $rheaders[0];

			preg_match( '{HTTP/\S*\s(\d{3})\S*\s(.*)}', $status_line, $match );

			$http_code = @(int) $match[1];
			$error_message = $match[2];

			if ( str_starts_with( $body, "\x1f\x8b\x08" ) ) {
				$body = gzdecode( $body );
			}
		}

		if ( ! empty( $error_message ) || $http_code !== 200 ) {
			if ( ( empty( $error_message ) || $error_message === 'GENERIC ERROR' ) && ! empty( $body ) ) {
				$body = @json_decode( $body, true );
				if ( ! empty( $body ) ) {
					$status = $this->getParam( $body, 'status', array() );
					$this->setLastStatus( $status );

					$http_code = $this->getParam( $status, 'code', $http_code );
					$error_message = trim( $this->getParam( $status, 'text' ) . ' - ' . $this->getParam( $status, 'error' ), " \t\n\r\0\x0B-" );
					$body = json_encode( $body );
				}
			}
			$this->setLastError( $error_message . " ($error_number - http code: $http_code)" );
		}

		return $body;
	}

	/**
	 * Prepares a file to be uploaded.
	 *
	 * @param string      $fileName name of the file to be uploaded
	 * @param string      $postname the name of the file in the upload data
	 * @param string|null $mimeType force the mime type for the file
	 *
	 * @return \CurlFile CURLFile to be used to upload a file (https://www.php.net/manual/en/class.curlfile.php)
	 *
	 * @example $this->prepareUpload( '/path/to/image/image.jpg', 'image.jpg' );
	 *
	 */
	function prepareUpload( $fileName, $postname, $mimeType = null ) {
		return new \CurlFile( $fileName, $mimeType ?: mime_content_type( $fileName ), str_replace( '"', '', $postname ) );
	}

	/**
	 * Helper that returns the value corresponding to a $key in the associative array $arr.
	 *
	 * @param array  $arr an array with keys to check
	 * @param string $key value to check
	 * @param mixed  $def default value to be returned if $key does not exists in $arr
	 *
	 * @return mixed the value of $key in $arr, $def otherwise
	 */
	function getParam( $arr, $key, $def = '' ) {
		return is_array( $arr ) && array_key_exists( $key, $arr ) ? $arr[ $key ] : $def;
	}

	private function getLogInURL() {
		$password = $this->getHash( $this->public_key );

		return $this->buildURL( 'users/login', array( 'username' => $this->username, 'password' => $password ) );
	}

	private function login( $force = false ) {
		if ( ! empty( $this->api_token ) || ( ! $force && ! empty( $this->session_id ) ) ) {
			return true;
		}

		$url = $this->getLogInURL();

		#try to login
		$result = $this->call( $url, null, false );

		if ( $result !== false ) {
			$session_id = $this->getParam( $result, 'session_id', '' );
			if ( ! empty( $session_id ) ) {
				$this->session_id = $session_id;

				return true;
			}
		}

		return false;
	}

	private function getHash( $value, $time = null ) {
		return hash_hmac( 'sha256', $value . ':' . ( $time ?: gmdate( "Y-m-d\TH:i\Z", time() ) ), $this->private_key );
	}

	public function getSessionID() {
		return $this->session_id;
	}

	/**
	 * Given an endpoint and parameters (optional) returns the URL to be called.
	 *
	 * @param string $endpoint   endpoint to be called
	 * @param array  $parameters associative array [param => value]
	 *
	 * @return string the URL to be called
	 */
	public function buildURL( string $endpoint, array $parameters = array() ): string {
		if ( ! is_array( $parameters ) ) {
			$parameters = array();
		}

		$encoded_parameters = http_build_query( array_merge( $parameters, $this->getSecurityParameters() ) );

		return self::$baseURL . $endpoint . ( ! empty( $encoded_parameters ) ? '?' . $encoded_parameters : '' );
	}

	private function getSecurityParameters(): array {
		$parameters = array();

		if ( $this->session_id !== '' ) {
			$parameters['_sid'] = $this->session_id;
			$parameters['_h'] = $this->getHash( $this->session_id );
		}

		return $parameters;
	}

	/**
	 * Set the return_errors flag.
	 *
	 * @param bool $bool set return error policy, if true the SDK will return errors, otherwise it will raise exceptions
	 */
	public function returnErrors( $bool ): void {
		$this->return_errors = $bool;
	}

	private function _wrapper( $callback, $retry = true, $force_login = false ) {
		try {
			if ( $this->login( $force_login ) ) {
				return $callback( $this );
			}

			throw new Exception( 'unable to log in' );
		} catch ( Throwable $t ) {
			$msg = $t->getMessage();
			if ( strpos( $msg, 'INVALID session' ) !== false ) {
				if ( $retry ) {
					return $this->_wrapper( $callback, false, true );
				}
				$msg = 'unable to log in';
			}

			throw new Exception( $msg );
		}
	}

	/**
	 * Automatically try to access the JotURL API, the callback is called if the access is successful.
	 *
	 * @param mixed $callback callback to be called on success
	 *
	 * @return mixed value returned by the $callback function, an exception is thrown on error
	 */
	function wrapper( $callback ) {
		return $this->_wrapper( $callback );
	}

	/**
	 * Set the base URL for API calls.
	 *
	 * @param string $baseURL base URL for API calls
	 *
	 * @return string
	 */
	static function setBaseURL( $baseURL ) {
		$oldBaseURL = self::$baseURL;

		self::$baseURL = $baseURL;

		return $oldBaseURL;
	}

	/**
	 * Get the base URL currently used for API calls.
	 *
	 * @return string
	 */
	static function getBaseURL() {
		return self::$baseURL;
	}
}
